Taken from the GranSim user's guide

> module Main where

> import Parallel

> main = print (parfib 20)

nfib :: Int -> Int
nfib 0 = 1
nfib 1 = 1
nfib n = nf1+nf2+1
         where nf1  = nfib (n-1)
               nf2  = nfib (n-2)

The straightforward idea for parallelising nfib is to start parallel
processes for computing both recursive calls.  This is expressed with
the par annotation, which `sparks' a parallel process for computing
its first argument and whose result is its second argument (the exact
operational semantics of par is discussed in more detail in chapter
section Parallelism Annotations.).  This gives the following parallel
program:

parfib :: Int -> Int
parfib 0 = 1
parfib 1 = 1
parfib n = nf2 `par` (nf1 `par` (nf1 + nf2 + 1))
           where nf1 = parfib (n-1)
                 nf2 = parfib (n-2)

The drawback of this program is the blocking of the main task on the
two created child-tasks.  Only when both child tasks have returned a
result, the main task can continue.  It is more efficient to have one
of the recursive calls computed by the main task and to spark only one
parallel process for the other recursive call.  In order to guarantee
that the main expression is evaluated in the right order (i.e. without
blocking the main task on the child task) the seq annotation is used:

> parfib :: Int -> Int
> parfib 0 = 1
> parfib 1 = 1
> parfib n = nf2 `par` (nf1 `seq` (nf1 + nf2 + 1))
>           where nf1 = parfib (n-1)
>                 nf2 = parfib (n-2)

The operational reading of this function is as follows: First spark a
parallel process for computing the expression parfib (n-2).  Then
evaluate the expression parfib (n-1).  Finally, evaluate the main
expression.  If the parallel process has not finished by this time the
main task will be blocked on nf2.  Otherwise it will just pick up the
result.
